package cytoPrefuse;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.plaf.metal.MetalButtonUI;

import cytoscape.Cytoscape;
import cytoscape.view.CytoscapeDesktop;

public class ControlPanel extends JPanel implements PropertyChangeListener, ListSelectionListener  {
	
	private static final long serialVersionUID = 1L;
	private Map displays = new HashMap();
	private float beta;
	private float labelDensity;
	private float  groupLabelSize;
	private boolean showHulls;
	private boolean showSeparators;
	private CytoPrefuse currentVis;
	private JTable groupsTable;
	private JButton addGroup;
	private JButton deleteGroup;
	private Color hullColor;
	
	private static ControlPanel controlPanel = new ControlPanel();
	 
	public static ControlPanel instance() {
		return controlPanel;
	}
	
	public void addDisplay (CytoPrefuse vis, String title){
		displays.put(title,vis);
	}
	public void removeDisplay (String title){
		displays.remove(title);		
	}
    
    public CytoPrefuse getCurrentVis(){
        return currentVis;
    }
	
	private ControlPanel(){
		Cytoscape.getDesktop().getSwingPropertyChangeSupport().addPropertyChangeListener(this);
		
		setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		setAlignmentX(Component.LEFT_ALIGNMENT);
	
	
		groupsTable = new JTable();
		groupsTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);		
		groupsTable.getSelectionModel().addListSelectionListener(this);
		groupsTable.addMouseListener(new MouseAdapter() {
			public void mouseClicked (MouseEvent e){
				int row = groupsTable.rowAtPoint(e.getPoint());
				if (row != -1){
					if (currentVis != null){
						currentVis.groupsManager.setCurrentGroup(row, e.isShiftDown());
					}
				}			
			}
		});
				
		JScrollPane scrollPane = new JScrollPane(groupsTable);
		scrollPane.setAlignmentX(Component.LEFT_ALIGNMENT);
		c.gridx = 0;
		c.gridy =0;
		c.weighty=1.0;
		c.weightx=1.0;
		c.anchor = GridBagConstraints.LINE_START;
		c.fill = GridBagConstraints.BOTH;
		c.gridwidth=2;
		add(scrollPane,c);
		
		JPanel buttonPanel = new JPanel();
		buttonPanel.setLayout( new BoxLayout(buttonPanel, BoxLayout.X_AXIS));
		
		
		addGroup = new JButton ("Add group");
		addGroup.setEnabled(false);
		addGroup.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				if (currentVis != null){
					currentVis.groupsManager.addGroup();
					groupsTable.setRowSelectionInterval(groupsTable.getRowCount()-1,groupsTable.getRowCount()-1);
					
				}
			}
		});

		deleteGroup = new JButton ("Delete group");
		deleteGroup.setEnabled(false);
		deleteGroup.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				int selectedGroup = groupsTable.getSelectedRow();
				GroupsManager groupsManager = (GroupsManager)groupsTable.getModel();
				groupsManager.removeGroup(selectedGroup);
				if (selectedGroup < groupsTable.getRowCount()){
					groupsTable.setRowSelectionInterval(selectedGroup, selectedGroup);
				}
			}
		});
		
		buttonPanel.add(addGroup);
		buttonPanel.add(Box.createHorizontalGlue());
		buttonPanel.add(deleteGroup);
		buttonPanel.setAlignmentX(Component.LEFT_ALIGNMENT);

		c.weighty=0.01;	
		c.gridy =1;
		c.fill = GridBagConstraints.HORIZONTAL;
		add (buttonPanel,c);
		
		beta = 1.0f;
		JLabel edgeControllerLabel = new JLabel ("Edge curviness", JLabel.CENTER);
		edgeControllerLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
		
		JSlider  edgeController  = new JSlider(JSlider.HORIZONTAL, 0, 10, 10);
		Hashtable labelTable = new Hashtable();
		
		labelTable.put (new Integer(0), new JLabel("Straight"));		
		labelTable.put (new Integer(10), new JLabel("Bundled"));
		edgeController.setLabelTable(labelTable);
		edgeController.setPaintLabels(true);
		edgeController.addChangeListener( new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				JSlider source = (JSlider)e.getSource();
	            beta = source.getValue() *0.1f;
	            Iterator displaysIt = displays.values().iterator();
	            while (displaysIt.hasNext()){
	            	CytoPrefuse cytoPrefuse = (CytoPrefuse)displaysIt.next();
	            	cytoPrefuse.edgeBundler.setBeta(beta);
	            	cytoPrefuse.getVisualization().run("color");
				    
	            } 	
			}
			
		});
		
		edgeController.setAlignmentX(Component.LEFT_ALIGNMENT);
		c.gridy=2;
		add (edgeControllerLabel,c);
		c.gridy=3;
		add (edgeController,c);
		
		
		labelDensity = 0;
		JLabel labelDensityLabel = new JLabel ("Label density", JLabel.CENTER);
		labelDensityLabel.setAlignmentX(Component.LEFT_ALIGNMENT);
		
		JSlider labelDensityController = new JSlider (JSlider.HORIZONTAL, 0, 100, 20);
		labelTable = new Hashtable();
		labelTable.put (new Integer(0), new JLabel ("Dense"));
		labelTable.put (new Integer(100), new JLabel ("Sparse"));
		labelDensityController.setLabelTable(labelTable);
		labelDensityController.setPaintLabels(true);
		labelDensityController.addChangeListener( new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				JSlider source = (JSlider)e.getSource();
	            labelDensity= source.getValue()/ 4000f;
	            Iterator displaysIt = displays.values().iterator();
	            while (displaysIt.hasNext()){
	            	CytoPrefuse cytoPrefuse = (CytoPrefuse)displaysIt.next();
	   		       cytoPrefuse.labelLayout.setExtraSpace(labelDensity);
	            }		     		
			}
			
		});
		labelDensityController.setAlignmentX(Component.LEFT_ALIGNMENT);
		add (Box.createRigidArea(new Dimension(0,20)));

		c.gridy=4;
		add (labelDensityLabel,c);
		c.gridy=5;
		add (labelDensityController,c);	
	
            
		groupLabelSize = 1.0f;
		JLabel labelDownstreamLabelSize = new JLabel ("Downstream label size", JLabel.CENTER);
		labelDownstreamLabelSize.setAlignmentX(Component.LEFT_ALIGNMENT);
		
		JSlider downstreamLabelSizeSlider = new JSlider (JSlider.HORIZONTAL, 0, 100, 40);
		labelTable = new Hashtable();
		labelTable.put (new Integer(0), new JLabel ("Off"));
		labelTable.put (new Integer(20), new JLabel ("Small"));
		labelTable.put (new Integer(100), new JLabel ("Large"));
		downstreamLabelSizeSlider.setLabelTable(labelTable);
		downstreamLabelSizeSlider.setPaintLabels(true);
		downstreamLabelSizeSlider.addChangeListener( new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				JSlider source = (JSlider)e.getSource();
				if (source.getValue() >= 20 ){
					groupLabelSize = source.getValue()/30.0f;
				}else{
					groupLabelSize = 0;
				}
				Iterator displaysIt = displays.values().iterator();
	            while (displaysIt.hasNext()){
	            	CytoPrefuse cytoPrefuse = (CytoPrefuse)displaysIt.next();
	   		        cytoPrefuse.labelLayout.setAggregateLabelScale(groupLabelSize);
	            }		     		
			}
			
		});
		downstreamLabelSizeSlider.setAlignmentX(Component.LEFT_ALIGNMENT);
	//	downstreamLabelSizeSlider.setForeground(Color.red);
	//	downstreamLabelSizeSlider.setBackground(Color.red);
		
		c.gridy=8;
		add (labelDownstreamLabelSize,c);
		c.gridy=9;
		add (downstreamLabelSizeSlider,c);	

		JCheckBox checkShowHulls = new JCheckBox("Show downstream clusters");
		checkShowHulls.addItemListener(new ItemListener() {

			public void itemStateChanged(ItemEvent e) {
				showHulls =e.getStateChange() == ItemEvent.SELECTED; 
			    Iterator displaysIt = displays.values().iterator();
	            while (displaysIt.hasNext()){
	            	CytoPrefuse cytoPrefuse = (CytoPrefuse)displaysIt.next();
	            	cytoPrefuse.setShowHulls(showHulls);
	            }
			}
		});
		checkShowHulls.setSelected(true);
		checkShowHulls.setAlignmentX(Component.LEFT_ALIGNMENT);		

		c.gridy=10;
		add (checkShowHulls,c);
		
		
		JLabel hullColorLabel = new JLabel("Cluster color");
		final JButton hullColorButton = new JButton("");
		hullColorButton.setUI(new MetalButtonUI());
		hullColorButton.setOpaque(true);
		hullColor = new Color(51,204,0);
		hullColorButton.setBackground(hullColor);
		hullColorButton.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {					
				Color selectedColor = JColorChooser.showDialog(null,"Choose group color", hullColor);
				if (selectedColor != null) {
					hullColor = selectedColor;
					hullColorButton.setBackground(hullColor);
					Iterator displaysIt = displays.values().iterator();
			        while (displaysIt.hasNext()){
			            	CytoPrefuse cytoPrefuse = (CytoPrefuse)displaysIt.next();
			    			cytoPrefuse.setHullColor(hullColor);			    			
			         }
				}				
			}		
		});
		c.gridx = 0;
		c.gridy =11;
		c.ipady = 20;
		c.weightx=0.5;
		c.gridwidth=1;
		c.fill = GridBagConstraints.NONE;
		hullColorLabel.setLabelFor(hullColorButton);
		add(hullColorLabel,c);
		c.gridx =1;
		
		add(hullColorButton,c);	
		
		
		JCheckBox checkShowSeparators = new JCheckBox("Show layer separators");
		checkShowSeparators.addItemListener(new ItemListener(){

			public void itemStateChanged(ItemEvent e) {
				showSeparators = e.getStateChange()==ItemEvent.SELECTED;
				Iterator displaysIt = displays.values().iterator();
		        while (displaysIt.hasNext()){
		           	CytoPrefuse cytoPrefuse = (CytoPrefuse)displaysIt.next();
		           	cytoPrefuse.setShowSeparators(showSeparators);
		        }
			}
			
		});
		checkShowSeparators.setSelected(true);
		checkShowSeparators.setAlignmentX(Component.LEFT_ALIGNMENT);
	
		c.fill =GridBagConstraints.HORIZONTAL;
		c.gridx=0;
		c.gridy=12;
		c.gridwidth=2;
		
		add(checkShowSeparators,c);
			
		
	}
	public boolean getShowHulls(){
		return showHulls;
	}

	public boolean getShowSeparators(){
		return showSeparators;
	}
	public float getBeta(){
		return beta;		
	}
	
	public float getDensity() {
		return labelDensity;
	}
	public Color getHullColor(){
		return hullColor;
	}
	
	public void propertyChange(PropertyChangeEvent e) {
		if (e.getPropertyName() == CytoscapeDesktop.NETWORK_VIEW_FOCUSED) {
			currentVis = (CytoPrefuse) displays.get(Cytoscape.getCurrentNetworkView().getTitle());
			if (currentVis != null){
				groupsTable.setModel(currentVis.groupsManager);
			}
			addGroup.setEnabled(currentVis != null);
			deleteGroup.setEnabled(groupsTable.getRowCount() > 0);
		}		else if (e.getPropertyName() == CytoscapeDesktop.NETWORK_VIEW_DESTROYED){
			currentVis = null;
			groupsTable.setModel(new GroupsManager());
			addGroup.setEnabled(false);
			deleteGroup.setEnabled(false);
		}
	}

	


	public void valueChanged(ListSelectionEvent e) {
		ListSelectionModel lsm = (ListSelectionModel) e.getSource();
		if (!lsm.isSelectionEmpty() && currentVis != null){
			currentVis.groupsManager.setCurrentGroup(lsm.getMaxSelectionIndex(), true);
		}
		deleteGroup.setEnabled(groupsTable.getRowCount()>0  && !lsm.isSelectionEmpty());
		
	}

	public float getGroupLabelSize() {
		return groupLabelSize;
	}

	

}
