package cytoPrefuse;

import java.awt.BasicStroke;
import java.util.Iterator;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import prefuse.Constants;
import prefuse.Display;
import prefuse.Visualization;
import prefuse.action.Action;
import prefuse.render.ExtendedShapeRenderer;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import cytoscape.CyNetwork;
import cytoscape.Cytoscape;
import cytoscape.visual.EdgeAppearance;
import cytoscape.visual.EdgeAppearanceCalculator;
import cytoscape.visual.GlobalAppearance;
import cytoscape.visual.GlobalAppearanceCalculator;
import cytoscape.visual.LineType;
import cytoscape.visual.NodeAppearance;
import cytoscape.visual.NodeAppearanceCalculator;
import cytoscape.visual.VisualStyle;

public class CytoscapeVisualMapper  extends Action implements ChangeListener  {
	    public static final byte RECT = (byte)0;
	    public static final byte ROUND_RECT = (byte)1;
	    public static final byte RECT_3D = (byte)2;
	    public static final byte TRAPEZOID = (byte)3;
	    public static final byte TRAPEZOID_2 = (byte)4;
	    public static final byte TRIANGLE = (byte)5;
	    public static final byte PARALLELOGRAM = (byte)6;
	    public static final byte DIAMOND = (byte)7;
	    public static final byte ELLIPSE = (byte)8;
	    public static final byte HEXAGON = (byte)9;
	    public static final byte OCTAGON = (byte)10;

	    private Visualization vis;
	    private CyNetwork cyNetwork;
	    public CytoscapeVisualMapper(Visualization vis, CyNetwork cyNetwork, ColorManager colorManager) {
	    	this.vis = vis;
	    	this.cyNetwork = cyNetwork;
	    	 Cytoscape.getVisualMappingManager().addChangeListener(this);
              
	   	  
	    }
	    public void applyVisualStyle() {
	    	//New visual mapping manager
			VisualStyle visualStyle = Cytoscape.getVisualMappingManager().getVisualStyle();
			NodeAppearanceCalculator nac = visualStyle.getNodeAppearanceCalculator();
			
			Iterator nodesIt = vis.getVisualGroup(CytoPrefuse.NODES).tuples();
			while (nodesIt.hasNext() ) {
				NodeItem nodeItem = (NodeItem)nodesIt.next();
				int cytoIndex = nodeItem.getInt(CytoPrefuse.CYTOSCAPE_ID);
				giny.model.Node cytoNode = cyNetwork.getNode(cytoIndex);
				NodeAppearance na = nac.calculateNodeAppearance(cytoNode, cyNetwork);
				nodeItem.set(CytoPrefuse.FILL_COLOR, na.getFillColor());
				nodeItem.set(CytoPrefuse.BORDER_COLOR, na.getBorderColor());
				LineType lineType = na.getBorderLineType();
				nodeItem.setStroke((BasicStroke) lineType.getStroke());
				nodeItem.setString(CytoPrefuse.LABEL, na.getLabel());
				nodeItem.set(CytoPrefuse.BASE_FONT,na.getFont());
				nodeItem.setSize(na.getWidth()/30);
				byte shape = na.getShape();
				switch (shape) {
					case RECT: nodeItem.setShape(Constants.SHAPE_RECTANGLE); break;
					case DIAMOND: nodeItem.setShape(Constants.SHAPE_DIAMOND); break;
					case ELLIPSE: nodeItem.setShape(Constants.SHAPE_ELLIPSE); break;
					case HEXAGON: nodeItem.setShape(Constants.SHAPE_HEXAGON); break;
					case TRIANGLE: nodeItem.setShape(Constants.SHAPE_TRIANGLE_UP); break;
					case PARALLELOGRAM: nodeItem.setShape(ExtendedShapeRenderer.SHAPE_PARALLELOGRAM);break;
					case ROUND_RECT: nodeItem.setShape(ExtendedShapeRenderer.SHAPE_ROUNDED_RECTANGLE); break;
					case OCTAGON: nodeItem.setShape(ExtendedShapeRenderer.SHAPE_OCTAGON); break;
					default:
						System.out.println ("Don't know how to handle node shape of " + shape);
				}
				nodeItem.set(CytoPrefuse.TEXT_COLOR, na.getLabelColor());
			}
			
			EdgeAppearanceCalculator eac = visualStyle.getEdgeAppearanceCalculator();
			Iterator edgesIt = vis.getVisualGroup(CytoPrefuse.EDGES).tuples();
			while (edgesIt.hasNext() ) {
				EdgeItem edgeItem = (EdgeItem)edgesIt.next();
				int cytoIndex = edgeItem.getInt(CytoPrefuse.CYTOSCAPE_ID);
				giny.model.Edge cytoEdge = cyNetwork.getEdge(cytoIndex);
				EdgeAppearance ea = eac.calculateEdgeAppearance(cytoEdge, cyNetwork);
				edgeItem.set(CytoPrefuse.FILL_COLOR, ea.getColor());
			    edgeItem.set(CytoPrefuse.LINE_TYPE, ea.getLineType());
			}
			
			GlobalAppearanceCalculator gac = visualStyle.getGlobalAppearanceCalculator();
			GlobalAppearance ga = gac.calculateGlobalAppearance(cyNetwork);
			Display d = vis.getDisplay(0);
			d.setBackground(ga.getBackgroundColor());
		//	colorManager.setCytoNodeSelected( ga.getNodeSelectionColor());
			//colorManager.setCytoEdgeSelected( ga.getEdgeSelectionColor());
			vis.run("aggLayout");
			vis.run("color");
	    }
	    public void stateChanged(ChangeEvent e) {
			applyVisualStyle();
		}
		public void run(double frac) {
			applyVisualStyle();
		}		
	                                  
}
