package cytoPrefuse;

import java.awt.Point;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.Vector;

import javax.swing.ProgressMonitor;
import javax.swing.SwingUtilities;

import prefuse.action.Action;
import prefuse.data.Edge;
import prefuse.data.Node;
import prefuse.data.tuple.TupleSet;
import prefuse.render.LineDividerRenderer;
import prefuse.util.GraphicsLib;
import prefuse.util.display.DisplayLib;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualGraph;
import prefuse.visual.VisualItem;

public class SearchGridLayout extends Action implements CerebralLayout {

        private static final int COOLING_PHASES = 40;
		private static final int NODE_EDGE_PENALTY = 25;
		private static final int EDGE_PENALTY = 5;
		private int gridHeight;
        private int gridWidth;
        
        private int rootV;
        
        int iteration;
        private List[] candidatePoints;
        private Node[][] nodeAtPosition;
        private int [] nodeCounts;
        private int[] layerBounds;
        
        private SpaceGrid spaceGrid;
        
        public static int SCALE = 60;
        private String cancelAction;
        private int iterationsWithoutImprovement;
        private int width;
        private Point neighbPoint = new Point();
        private Random generator = new Random( );    
        private Point[] nodepos;
        private Point nodeOrigSave = new Point();
        private boolean firstRun;       
        private VisualGraph graph;
        private int numLayers;
        private LineDividerRenderer lineDividerRenderer;
        private double temperature;
        private boolean minorImprovements;
        private boolean quitEarly;
        private ProgressMonitor progressMonitor;
        private long startTime;
        
        
        
        //For grid scoring evaluation
        public  SearchGridLayout(VisualGraph inGraph, String inCancelAction, int numLayers, LineDividerRenderer lineDividerRenderer) {
                
                 
            cancelAction = inCancelAction;
        
            graph = inGraph;
            nodepos = new Point[graph.getNodeCount()+1];           
            rootV = (int)Math.sqrt(graph.getNodeCount());
            width =(int) (2*rootV);                                
            this.numLayers = numLayers;
            this.lineDividerRenderer = lineDividerRenderer;
        	firstRun = true;
        }
        

        
        public boolean pointOccupied(int x, int y){
        	return nodeAtPosition[x/SCALE][y/SCALE] != null;
        }
        public NodeItem getNodeAtPosition (int x, int y) {
         	return (NodeItem)nodeAtPosition[x/SCALE][y/SCALE];
            
        }
        private void checkNodeAtPos() {
        	Set nodesSeen = new HashSet();
        	for (int x = 0; x < nodeAtPosition.length; x++){
        		for (int y = 0; y <nodeAtPosition[0].length; y++){
        			Node n = nodeAtPosition[x][y];
        			if (n!=null){
        				if (nodesSeen.contains(n)){
        					System.out.println("NODESSEEN inconsisten");
        				}else {
        					nodesSeen.add(n);
        				}
        			}
        		}
        	}

        	Set uniqueNodeLoc = new HashSet();
            
        	for (int i =0; i < nodepos.length; i++){
        		Point p = nodepos[i];
	       		if (p != null){	        		
	        		Integer hash = new Integer (p.x*10000+p.y);
		        	if (hash.intValue() != 0){
		        		if (uniqueNodeLoc.contains(hash)){
		        			System.out.println("nodepos contains two different points that map to the same location");
		        			
		        		}
		        		uniqueNodeLoc.add(hash);
		        	}
	        	}
	        }
        }
        
        public void initialize(boolean assignNodePositions) {
            startTime = System.currentTimeMillis();
            System.out.println ("Starting layout.");
        	for (int i =0; i < nodepos.length; i++){
        		nodepos[i] = new Point();
        	}
            iteration = 1;
            iterationsWithoutImprovement = 0;
            //add 1 for  the downstream layer
            nodeCounts = new int [numLayers + 1];
            Arrays.fill(nodeCounts, 0);

            TupleSet nodes = graph.getNodes();
            
            int anyLayerCount =0;
            Iterator nodesIt = nodes.tuples();
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();

                if (node.getBoolean(CytoPrefuse.ANY_LAYER) && ! node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
                	anyLayerCount++;
                } else{
                	if (node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
                		  nodeCounts[numLayers]++;
	                }else{
		                int bioCategory = node.getInt(CytoPrefuse.BIO_CATEGORY);
		                nodeCounts[bioCategory]++;
	                }
                }
            }
            
        
            gridHeight = (int) (2* rootV);            
            int gridWidth = gridHeight;
                                         
            layerBounds = new int [numLayers +1];
            int bottom = 0;
            int top = 0;
              
            float avgAnyLayersPerLayer = anyLayerCount / (numLayers+1);
            for (int i = 0; i < numLayers + 1; i++){
                bottom = top;                           
                       
                top += (int) (((nodeCounts[i]+ avgAnyLayersPerLayer )/(float)nodes.getTupleCount()) * gridWidth) ; 
                if (top - bottom == 0){
                    top++;
                    gridHeight++;
                }
                if (top - bottom == 1) {
                    top++;
                    gridHeight++;
                }
                if (top - bottom == 2) {
                    top++;
                    gridHeight++;
                }
                layerBounds[i] = top;
            }
            System.out.println ("Using a grid size of [" +gridWidth + ":" + gridHeight + "]");
                    
            layerBounds[layerBounds.length -1 ] =gridHeight;
            lineDividerRenderer.setDividers(layerBounds); 
            
            int rightBorder = width;                   
               
            candidatePoints = new ArrayList [numLayers +1];               
            for (int i=0; i <numLayers +1; i++){
                candidatePoints[i] = new ArrayList();
            }

            bottom = 0;
            top = 0;
            for (int i = 0; i < nodeCounts.length; i++){
                bottom = top;
                top = layerBounds[i];
                for (int j=0; j < rightBorder; j++){
                    for (int k=bottom+1; k < top; k++){
                        Point p = new Point (j,k);
                        candidatePoints[i].add( (p));
                    }
                }
            }    
       
            nodesIt = nodes.tuples();	            
            int maxGridPoint = Integer.MIN_VALUE;
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();      	                
                int x = (int) node.getEndX();
                if (x/SCALE > maxGridPoint){
                	maxGridPoint = x/SCALE;
                }
                int nx =nodepos[node.getRow()].x;
                if (nx > maxGridPoint){
                	maxGridPoint = nx;
                }
            }
            maxGridPoint = (maxGridPoint > gridHeight)? maxGridPoint : gridHeight;
            
            initializeNodeAtPosition (maxGridPoint , gridHeight);
            if (assignNodePositions)
                assignNodePositions (nodes, 0, rightBorder, candidatePoints);     
            initializeSpaceGrid();

			 //We want the initial temperature to be the temperature of our current layout			           
             nodesIt = nodes.tuples();               
             while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();                          
                temperature += scoreNode(node);
             }
             temperature/= nodes.getTupleCount();
             
			 minorImprovements=false;
			 quitEarly = false;
        }

        public int[] getLayerBounds() {
        	return layerBounds;
        }
        
        public void run(double frac) {
            CytoPrefuse cytoPrefuse = (CytoPrefuse) m_vis.getDisplay(0);                                 
            
        	if (firstRun) {
                Rectangle2D bounds = cytoPrefuse.getVisualization().getBounds(CytoPrefuse.NODES);
                GraphicsLib.expand(bounds,
                        50 + (int) (1 / cytoPrefuse.getScale()));
                DisplayLib.fitViewToBounds(m_vis.getDisplay(0), bounds, 1000);        	       
        		firstRun = false;
        	}
        	improveLayout ( candidatePoints); 
        
     		m_vis.run("color");
      
            if (iteration == COOLING_PHASES || quitEarly) {      
            	System.out.println ("Layout finished.  Total iterations: " + iteration);
            	System.out.println ("Layout time. " + (System.currentTimeMillis() - startTime));
             	SwingUtilities.invokeLater(new Runnable(){

					public void run() {
						progressMonitor.close();
					}
				
				});
                //We're done.
                m_vis.cancel(cancelAction);
  
                EdgeBundler edgeBundler = cytoPrefuse.edgeBundler;
                if (edgeBundler != null){
                    edgeBundler.recomputeLocations();
                }
                cytoPrefuse.setCytoscapePostionsFromLayout(graph.getNodes());
                
                m_vis.run("aggLayout");             
                Rectangle2D bounds = cytoPrefuse.getVisualization().getBounds(CytoPrefuse.NODES);
                GraphicsLib.expand(bounds,
                        50 + (int) (1 / cytoPrefuse.getScale()));
                DisplayLib.fitViewToBounds(m_vis.getDisplay(0), bounds, 1000);                 
                m_vis.invalidateAll();
                m_vis.run("repaint"); 
            }else{
            	temperature = temperature * 0.6;
            }
        }
        
	
		private void initializeSpaceGrid() {
		
		    //the spacegrid needs to be big enough for any node to be tracked,
		    //not just the nodes we're laying out (for the nodes we are laying
		    //out could join to nodes outside the region.)
		    int maxX = Integer.MIN_VALUE;
		    int maxY = Integer.MIN_VALUE;
		    TupleSet nodes = graph.getNodes( );
		    Iterator nodesIt = nodes.tuples();
		    while (nodesIt.hasNext()){
		        NodeItem node = (NodeItem)nodesIt.next();
		        Point p = nodepos[node.getRow()];
		        if (p.x > maxX){
		            maxX = p.x;
		        }
		        if (p.y > maxY){
		            maxY = p.y;
		        }
		    }               
		    gridWidth = maxX < 0 + width ? 0+ width : maxX;
		    gridHeight = maxY < gridHeight ? gridHeight : maxY;
		    spaceGrid = new SpaceGrid (gridWidth + 1, gridHeight + 1, graph.getNodeCount(), this);
		        
		        
		    Iterator iter = graph.getNodes().tuples();
		    Set edgesAdded = new HashSet();
		    while ( iter.hasNext() ) {
		        NodeItem node = (NodeItem)iter.next();
		        spaceGrid.addNode(node);
		        Iterator edgesIt = node.edges();
		        while (edgesIt.hasNext()){
		            Edge edge = (Edge)edgesIt.next();
		            if (!edgesAdded.contains(edge)){
		                spaceGrid.addEdge(edge);
		                edgesAdded.add(edge);
		            }
		        }
		    }
		 }
        
        private void initializeNodeAtPosition(int x, int y) {
            nodeAtPosition = new Node[x+1][y+1];
            for (int xi=0; xi <= x; xi++) {
                Arrays.fill(nodeAtPosition[xi],null);
            }
        }

        public final Point getPositionOfNode (Node n){
        	return nodepos[n.getRow()];
        }
        

        private void improveLayout(List[] points) {
                
            final long startTime = System.currentTimeMillis();
            iteration ++;
            int escapeMinima =0;
            int improveCount = 0;
            
            for (int i=0; i < 30 * graph.getNodeCount(); i++){
            	
            	//pick a node at random                  
            	NodeItem node = (NodeItem) graph.getNode(generator.nextInt(graph.getNodeCount()));
            	
            	Point trialPoint;
                do {
	                //and a random spot in the grid                
	                int layer =0;
	                if (node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
	                	layer = numLayers;
	                }else {
	                	layer =  node.getInt(CytoPrefuse.BIO_CATEGORY);                    
	                }
	                boolean anyLayer =node.getBoolean(CytoPrefuse.ANY_LAYER) && !node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE); 
	                if (anyLayer){
	                   layer = generator.nextInt(numLayers + 1);
	                }
	                int sample = generator.nextInt(points[layer].size());
	                trialPoint =  (Point) points[layer].get(sample);
                }while (!pointIsEmpty(trialPoint));   
                
                int improvement = 0;
                removeNodeFromSpaceGrid(node);
                int oldScore = scoreNode (node);
                nodeOrigSave.x = nodepos[node.getRow() ].x;
                nodeOrigSave.y = nodepos[node.getRow() ].y;
                    
            	moveNodeToEmptyPoint(node, trialPoint, false);
                improvement = oldScore - scoreNode (node);
                boolean acceptChange = false;
                if (improvement > 0){
                	acceptChange = true;
                }
                else {
                	//SImulated annealing test
                	if (Math.exp (improvement/temperature) > Math.random()){
                		//	System.out.println ("Ratio " + (improvement/temperature) + " :exp " +Math.exp (improvement/temperature)); 
                		acceptChange = true;
                		escapeMinima ++;
                	}
                }
                if (acceptChange){
                	  iterationsWithoutImprovement =0;
                	  improveCount++;
                      setDisplayFromNodepos(node);                         
                }else{
                	moveNodeToEmptyPoint(node, nodeOrigSave,false);
                    iterationsWithoutImprovement++;
                }
                addNodeToSpaceGrid(node);
                
                if ((i % 100) == 0){
                	if (progressMonitor.isCanceled()){
                		quitEarly = true;
                		break;
                	}
                }
            }
            
            if (improveCount / (float) escapeMinima < 1.4f){
            	if (minorImprovements){
            		quitEarly = true;
            	}else {
            		minorImprovements = true;
            	}
            }else{
            	minorImprovements = false;            	
            }
            
            System.out.println ("Temperature: " + temperature + " EscapeMinima: "+ escapeMinima + " Improvements:" + improveCount);
            
          	SwingUtilities.invokeLater(new Runnable(){

					public void run() {
				        progressMonitor.setProgress((int) (iteration/40.0f * 1000));	
				        String message = "Estimated time left ";
				  	    long estimatedTimeLeft = (COOLING_PHASES - iteration) * (System.currentTimeMillis() - startTime)	/ 1000;
						if (estimatedTimeLeft > 90) {
							message += estimatedTimeLeft / 60 + " minutes.";
						}else {
							message += " < " + ((estimatedTimeLeft  /15)+1) * 15 + " seconds.";
						}
				        progressMonitor.setNote(message);
					}
				
				});
			
          
   }
        

        private void addNodeToSpaceGrid(Node node) {
            spaceGrid.addNode((NodeItem) node);             
            Iterator edges = node.edges();
            while (edges.hasNext()) {
                Edge e = (Edge) edges.next();
                spaceGrid.addEdge(e);
            }               
        }
        

               
        private void removeNodeFromSpaceGrid(Node node) {
            Iterator edges = node.edges();
            while (edges.hasNext()) {
                    Edge e = (Edge) edges.next();
                    spaceGrid.deleteEdge(e);
            }
            spaceGrid.deleteNode((NodeItem) node);
        }

		
        private int scoreNode(NodeItem node) {        	
            int score = 0;
            
             
             if ( node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)) {                        
                score = countCrossings(node);          
               score += countNeighbourNodeDistance (node);
                score += 50*EDGE_PENALTY *  countFunctionNeighbourNodeDistance (node);
                score += 50*EDGE_PENALTY* countOtherFunctionInNeighbourhood(node);
            }else{
               score = countCrossings(node);                     
               score += countNeighbourNodeDistance (node);                              
               score += countFunctionNeighbourNodeDistance(node);   
            }
            
           //score = countCrossings(node);
            return score;
        }

		private int countNeighbourNodeDistance (NodeItem node, Iterator neighbs) {
            int score =0;
            Point nodePos = nodepos[node.getRow()];
            while (neighbs.hasNext()){
                Node neighbour = (Node) neighbs.next();
                Point neighbPos = nodepos[neighbour.getRow()];
                int increment =Math.abs ( neighbPos.x - nodePos.x)  +Math.abs( neighbPos.y - nodePos.y);        
                score += increment;            
            }
            return score;
        }

        private int countFunctionNeighbourNodeDistance(NodeItem node) {
          
            Vector functionNeighbourList = (Vector) node.get(CytoPrefuse.FUNCTION_NEIGHBOURS);
            int score =0;
            if (functionNeighbourList != null) {
                score = countNeighbourNodeDistance(node, functionNeighbourList.iterator());             
            }
            return score;           
        }

        private int countOtherFunctionInNeighbourhood(NodeItem node) {
            int score = 0;
            boolean functionUnknown = node.getBoolean(CytoPrefuse.BIO_FUNCTION_UNKNOWN);           
            int bioFunction = node.getInt(CytoPrefuse.BIO_FUNCTION);
            Point pos = nodepos[node.getRow()];
            for (int x = -1; x <=1; x++) {
                for (int y =-1; y <=1; y++) {
                    neighbPoint.x = pos.x + x;
                    neighbPoint.y = pos.y + y;
                    if ( neighbPoint.x>= 0 && neighbPoint.x < nodeAtPosition.length &&
                         neighbPoint.y >= 0 && neighbPoint.y < nodeAtPosition[0].length &&
                         (x!=0 || y!=0) ){
                        Node neighbour = nodeAtPosition[neighbPoint.x][neighbPoint.y];
                        if (neighbour != null) {
                            int neighbBioFunction = neighbour.getInt(CytoPrefuse.BIO_FUNCTION);
                            boolean neighbFunctionUnknown = neighbour.getBoolean(CytoPrefuse.BIO_FUNCTION_UNKNOWN);
                            if (!functionUnknown && !neighbFunctionUnknown){
                            	if (neighbBioFunction != bioFunction ) {
                            		score ++;
                            	}
                            }else if (functionUnknown ^ neighbFunctionUnknown){
                            	score++;
                            }
                        }
                    }
                }
            }
            return score;

        }

        private int countNeighbourNodeDistance(NodeItem node) {
                return countNeighbourNodeDistance(node, node.neighbors());
        }

        private int getLayer(int yCoord) {
            int testY = yCoord;
            for (int i =0; i < layerBounds.length; i++){
                if (testY < layerBounds[i])
                    return i;
            }
            return -1;
        }

        private int countCrossings(NodeItem node) {
            int edgeCrossings = 0;
            Iterator edgesIt = node.edges();
            while (edgesIt.hasNext()){      
                Edge e = (Edge) edgesIt.next();
                edgeCrossings += spaceGrid.countCrossings (e, EDGE_PENALTY, NODE_EDGE_PENALTY);                
            }               
           
            int thisNodeEdgeCrossings = spaceGrid.countCrossings(nodepos[node.getRow()], NODE_EDGE_PENALTY);
           return edgeCrossings + thisNodeEdgeCrossings;
           
      }
        
        private void assignNodePositions(TupleSet nodes, int leftBorder, int rightBorder, List[] points) {
            Random generator = new Random();
                    
            Set unplaceableNodes = new HashSet();
          
            Iterator nodesIt = nodes.tuples();		            
            //Place the nodes to their closest gridpoint
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();      
                
                boolean couldSetPosition = setNodeposFromDisplay(node, Integer.MIN_VALUE, Integer.MAX_VALUE);
                if (!couldSetPosition){
                    unplaceableNodes.add(node);
                }                       
            }
            //And place the filtered nodes within the boundaries
            nodesIt = nodes.tuples();
            
            //Place the nodes to their closest gridpoint
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();      
                
                boolean couldSetPosition = setNodeposFromDisplay(node, leftBorder, rightBorder);
                if (!couldSetPosition){
                    unplaceableNodes.add(node);
                }
            }
            
            //nodes that weren't previously on the grid get placed at random
            
            nodesIt = unplaceableNodes.iterator();
            while (nodesIt.hasNext()){
                NodeItem node = (NodeItem) nodesIt.next();                      
                int layer = getLayer(generator, node);
                int randPos;
                Point pCandidate;
                do{
                    randPos = generator.nextInt( points[layer].size());             
                    pCandidate = (Point) points[layer].get(randPos);
                }while (!pointIsEmpty(pCandidate));
                
                Point p = new Point (pCandidate.x, pCandidate.y);
                nodepos[node.getRow()]=p;
	            nodeAtPosition[p.x][p.y] = node;
                setDisplayFromNodepos(node);
            }
	     }

        private void setDisplayFromNodepos(NodeItem node){
        	Point p = nodepos[node.getRow()];
       		 node.setX(p.x*SCALE);
       		 node.setY(p.y *SCALE);
             node.setEndX(p.x *SCALE);
             node.setEndY(p.y *SCALE);                     
        }
		private boolean setNodeposFromDisplay(NodeItem node, int left, int right) {
			int x = (int) node.getEndX();
			int y = (int) node.getEndY();
           
			Point p = nodepos[node.getRow()];
			p.x = x/SCALE;
			p.y = y/SCALE;
			if (p.x < 0 || p.x > nodeAtPosition.length-1 ||
					p.y < 0 || p.y > nodeAtPosition[0].length -1 ){
				return false;
			}
			boolean couldSetPosition = false;
			if (x%SCALE ==0 && y%SCALE == 0){
				if (p.x > left && p.x < right){
				    if (pointIsEmpty(p)){
				        //see if it's a legal layer for the node
				        int layer = getLayer (p.y);
				        boolean legal = false;
				        if (node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
				        	legal = (layer == numLayers);
				        }else {					        	
					        int bioCategory = node.getInt(CytoPrefuse.BIO_CATEGORY);
					        legal =node.getBoolean(CytoPrefuse.ANY_LAYER) ||
					        	   bioCategory == (layer);
				        }
				        if ( legal ) {
				            couldSetPosition =true;
				            nodeAtPosition[p.x][p.y] = node;            
				        }else {
				        	//	System.out.println("Illegal layer");
				        }
				    }else {
	//			    	System.out.println("Point wasn't empty");
				    }			
				}else {
		//			System.out.println("Outside bounds");
				}
			}
			return couldSetPosition;
		}

        private void moveNodeToEmptyPoint(Node node, Point p, boolean updateGrid) {
            if (updateGrid)
                removeNodeFromSpaceGrid(node);
            
            Point oldPos =  nodepos[node.getRow()];
            nodeAtPosition[oldPos.x][oldPos.y] = null;
                
            oldPos.x = p.x;
            oldPos.y = p.y;
                
            nodeAtPosition[p.x][p.y] = node;        
            if (updateGrid)
                addNodeToSpaceGrid(node);
        }
        
      


		private boolean pointIsEmpty(Point p) {
			if (nodeAtPosition[p.x][p.y] != null)
				return false;
			else{
				//make sure it's not a layer boundary
				for (int i=0; i < layerBounds.length; i++) {
					if (p.y == layerBounds[i])
						return false;					
				}
				return true;
			}
        }

        private int getLayer(Random generator, Node node) {
            int layer =0;
            if (node.getBoolean(CytoPrefuse.ANY_LAYER) && !node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE)){
                    int yPoint= generator.nextInt (gridHeight);
                    layer = getLayer(yPoint);
            }else {
            	if (node.getBoolean(CytoPrefuse.DOWNSTREAM_GENE))
            		layer= numLayers;
            	else {
            		layer = node.getInt(CytoPrefuse.BIO_CATEGORY);
            	}
            }
            return layer;
        }

		public int getGridWidth() {
			return gridWidth;
		}
		
		
		public int getGridHeight() {
			return gridHeight;
		}

		public Point pointLocation(VisualItem item) {
			return nodepos[item.getRow()];
		}

		public void setNodeAtLocation(VisualItem item, int unscaledx, int unscaledy) {
			Point scaledPoint = new Point(unscaledx/SCALE, unscaledy/SCALE);
			moveNodeToEmptyPoint((Node) item, scaledPoint, false);
		}

		public void setProgressMonitor(ProgressMonitor progressMonitor) {
			this.progressMonitor = progressMonitor;
		}

     
}